#include "checksum.h"

#include "map.h"
#include "fileio.h"
#include "clientprot.h"


//////// Simple Transfer checksum ////////
// Assembly rips by Coconut Emulator

BYTE simpleChecksum(void *buffer, Uint32 len)
{
	BYTE c = 0;

	for (Uint32 i = 0; i < len; ++i)
	{
		c ^= ((BYTE*)buffer)[i];
	}

	return c;
}


//////// File checksum ////////
// This section made possible by Snrrrub

Uint32 getFileChecksum(BYTE *buffer, long *dictionary, Uint32 len)
{
	Uint32 Index = 0,
		   Key   = Index - 1;

	for (Uint32 i = 0; i < len; ++i)
	{
		Index	= dictionary[(Key & 255) ^ buffer[i]];
		Key		= (Key >> 8) ^ Index;
	}

	return ~Key;
}

Uint32 getFileChecksum(char *fileName, long *dictionary)
{
	Uint32 chk = UNASSIGNED;

	fileAccess fa;
	if (fa.open(fileName, true))
	{
		Uint32 len = fa.getLength();
		BYTE *file = (BYTE*)fa.readStuff(0, len);
		chk = getFileChecksum(file, dictionary, len);
		delete []file;
		file = NULL;
	}

	return chk;
}

void generate4(long *offset, long key)
{
	offset[0] = key;
	offset[1] = key ^ G4_MODIFIER;
	offset[2] = key ^= (G4_MODIFIER << 1);
	offset[3] = key ^ G4_MODIFIER;
}

void generate16(long *offset, long key)
{
	generate4(offset, key);
	generate4(offset + 4, key  ^ G16_MODIFIER);
	generate4(offset + 8, key  ^= (G16_MODIFIER << 1));
	generate4(offset + 12, key ^ G16_MODIFIER);
}

void generate64(long *offset, long key)
{
	generate16(offset, key);
	generate16(offset + 16, key ^ G64_MODIFIER);
	generate16(offset + 32, key ^= (G64_MODIFIER << 1));
	generate16(offset + 48, key ^ G64_MODIFIER);
}

void generateDictionary(long *offset, long key)
{
	generate64(offset, key);
	generate64(offset + 64, key  ^ G256_MODIFIER);
	generate64(offset + 128, key ^= (G256_MODIFIER << 1));
	generate64(offset + 192, key ^ G256_MODIFIER);
}


//////// Security checksum ////////
//  Supreme thanks go to Coconut Emulator and Snrrrub
//  for providing the assembly rips.

/*	 Doesn't work, it's on ice =/
Uint32 generateLevelChecksum(Uint32 key, char * mapData)
{
	Sint32 okey = key;							// store
	Sint32 x = key % 31;						// EDX +/- 0-30
	Sint32 y = key % 32;						// ECX +/- 0-31
	char *linear = &mapData[getLinear(x, y)];	// ESI get address
	Sint32 rx = 1024 - x;						// EAX (1023 - x) + 1
	Sint32 ry = (1055 - y) >> 5;				// cnt (1023 - y) / 32 + 1

	while (ry-- > 0)
	{
		char *target = rx + linear;
		char *off = linear;

		while (off < target)
		{
			BYTE tile = *off;
			if (tile && (tile < vieNormalEnd) || (tile == vieSafeZone))
			{
				key += okey ^ tile;
			}

			off += 31;		// skip 31 tiles
		}

		linear += 0x8000;	// skip 32 lines
	}

	return key;
}
*/

Uint32 generateLevelChecksum(Uint32 key, char *mapData)
{	// Snrrrub's literal version
	LONG EAX, ECX, ESI, EDX;
	LONG cnt;

	if (key & 0x80000000)
	{	// key is negative
		ECX = -((-signed(key)) & 0x1F);

		if (ECX >= 1024)
			return key;
	}
	else
	{	// key is positive
		ECX = key & 31;
	}

	EDX = key % 31;

	ESI = (ECX << 0x0A) + (LONG)mapData + EDX;
	EAX = 1024 - EDX;
	cnt = (1024 + 31 - ECX) >> 5;
	mapData = (char *)EAX;

	LONG original_key = key;

	for (; cnt > 0; --cnt, ESI += 0x8000)
	{
		EDX = EAX + ESI;
		ECX = ESI;

		if (ESI >= EDX)
			continue;

		while (ECX < EDX)
		{
			BYTE byte = (*((char *)ECX));

			if ((byte) && (byte < 0xA1 || byte == 0xAB))
				key += original_key ^ byte;

			ECX += 31;
		}
		EAX = (LONG)mapData;
	}

	return key;
}

Uint32 generateParameterChecksum(Uint32 key, Uint32 *settings)
{
	Uint32 i, c = 0;

	for (i = 0; i < 357; ++i)
	{
		c += (settings[i] ^ key);
	}

	return c;
}

Uint32 generatePartialEXE(Uint32 key, char *buffer, Uint32 len)
{
	Uint32 data,
		   partialKey = 0,
		   offset = 0;

	while (len > 4)
	{
		data = getLong(buffer, offset);

		if ((len & 0x7FF) == 0)
		{
			partialKey += (data | key);
		}
		else
		{
			partialKey ^= (data ^ key);
		}

		len -= 3;
		offset += 3;
	}

	delete []buffer;
	buffer = NULL;

	return partialKey;
}

Uint32 generateEXEChecksum(Uint32 key, char *fileName)
{
	fileAccess fa;
	char *fd;

	if (fa.open(fileName, true))
	{
		Uint32 c = 0;

		c += generatePartialEXE(key, fd = fa.readStuff(0x00006940, 0x0000D5A0), 0x0000D5A0);
		c += generatePartialEXE(key, fd = fa.readStuff(0x00000400, 0x00002480), 0x00002480);
		c += generatePartialEXE(key, fd = fa.readStuff(0x0001CAE0, 0x00006F70), 0x00006F70);
		c += generatePartialEXE(key, fd = fa.readStuff(0x00042E70, 0x000021E0), 0x000021E0);
		c += generatePartialEXE(key, fd = fa.readStuff(0x000305A0, 0x0000D3C0), 0x0000D3C0);
		c += generatePartialEXE(key, fd = fa.readStuff(0x0002A5B0, 0x00005FE0), 0x00005FE0);
		c += generatePartialEXE(key, fd = fa.readStuff(0x00003960, 0x00001C90), 0x00001C90);
		c += generatePartialEXE(key, fd = fa.readStuff(0x00013EF0, 0x00008BE0), 0x00008BE0);
		c += generatePartialEXE(key, fd = fa.readStuff(0x00005750, 0x000000E0), 0x000000E0);

		return c;
	}
	else
	{
		return -1;
	}
}
